@@ -19,6 +19,7 @@ from typing import List, Tuple, Optional, Union, Type
 from collections import OrderedDict
 import sys
 import zeep
+import ast
 sys.path.append("..")
 try:
     from custom_exceptions.salesforce_exceptions import *
@@ -704,6 +705,11 @@ class Sf:
         
         return result_dict
     
+    async def async_set_user_passwords(self, user_id_password_list_dict: dict) -> dict:
+        async_set_user_password = async_wrap(self.set_user_password)
+        results = await asyncio.gather(*[async_set_user_password(user_id, password) for user_id, password in user_id_password_list_dict.items()])
+        results = {result['user_id']: result['message'] for result in results}
+        return results
 
     async def entity_access_bulk_query(self, entity_types: list):
         async_entity_access_query = async_wrap(self.entity_access_query)
@@ -814,6 +820,106 @@ class Sf:
         else:
             return account_page_layouts_bulk
 
+    async def field_permission_user_check(self, objects: list[str], permissions: list[str], users=None) -> pd.DataFrame:
+        # Similar to the field_permission_check function, but this one will return a list of users that have the specified permission on the specified object
+        
+        # this is to avoid having a mutable default argument
+        users = [] if users is None else users
+        
+        object_query = "','".join(objects)
+        permission_query = ' OR '.join([f"({permission} = true)" for permission in permissions])
+
+        relevant_ps_and_profiles = transform_sf_result_set_rec(self.sf.query_all(f"""
+                                                SELECT ParentId,
+                                                        Field,
+                                                        PermissionsEdit,
+                                                        PermissionsRead
+                                                FROM FieldPermissions
+                                                WHERE SObjectType IN ('{object_query}') AND
+                                                ({permission_query})
+                                                """)['records'])
+
+        relevant_ps_and_profiles['Object API Name'] = relevant_ps_and_profiles['FieldPermissions.Field'].apply(lambda row: row.split('.')[0])
+        relevant_ps_and_profiles['Field API Name'] = relevant_ps_and_profiles['FieldPermissions.Field'].apply(lambda row: row.split('.')[1])
+        relevant_ps_and_profiles.drop(columns=['FieldPermissions.Field'], inplace=True)
+
+        user_query = (" AND Assignee.UserName IN ('" + "','".join(users) + "')") if len(users) > 0 else ""
+
+        relevant_ps_and_profiles_list = "','".join(relevant_ps_and_profiles['FieldPermissions.ParentId'].unique().tolist())
+
+        data = transform_sf_result_set_rec(self.sf.query_all(f"""SELECT Assignee.Id, Assignee.Name,Assignee.IsActive, 
+                                                                Assignee.UserName, PermissionSet.Id, 
+                                                                PermissionSet.isOwnedByProfile, PermissionSet.Profile.Name, PermissionSet.Label
+                                                                FROM PermissionSetAssignment
+                                                                WHERE PermissionSetId
+                                                                IN ('{relevant_ps_and_profiles_list}')
+                                                                {user_query} AND Assignee.IsActive = TRUE""")['records'])
+        
+        if len(data) == 0:
+            return pd.DataFrame(columns=['UserName', 'Profile', 'Field API Name', 'Object API Name','Permission', 'Permission Set'])
+        
+        # Retrieve Profile FLS data
+        profiles = data[data['PermissionSet.IsOwnedByProfile']]['Profile.Name'].unique().tolist()
+        profiles = ['Admin' if x == 'System Administrator' else x for x in profiles]
+
+
+        async_metadata_read = async_wrap(self.metadata_read)
+        profile_metadata_tasks = {}
+        profile_metadata_dict = {}
+
+        for profile in profiles:
+                profile_metadata_tasks[profile] = asyncio.create_task(async_metadata_read('Profile', profile))
+
+
+        task_result_list = await asyncio.gather(*profile_metadata_tasks.values())
+
+
+        for profile in profiles:
+                profile_metadata_dict["System Administrator" if profile == 'Admin' else profile] = pd.DataFrame(task_result_list[profiles.index(profile)]['fieldPermissions'][0])
+    
+        data = data.merge(relevant_ps_and_profiles, left_on='PermissionSet.Id', right_on='FieldPermissions.ParentId', how='left').copy()
+        if (len(data) > 0):
+                
+                profile_violations = data[data['PermissionSet.IsOwnedByProfile']== True][['User.Username', 'Profile.Name']].drop_duplicates().copy()
+                # profile_violations = data[data['PermissionSet.IsOwnedByProfile']== True][['User.Username', 'Profile.Name','Field API Name']]
+                profile_metadata_pd = pd.concat(profile_metadata_dict.values(), keys=profile_metadata_dict.keys(), names=['Profile', 'Index']).reset_index(level=1, drop=True).reset_index()
+                profile_metadata_pd.rename(columns={'Profile': 'Profile.Name'}, inplace=True)
+                profile_metadata_pd['Field API Name'] = profile_metadata_pd['field'].str.split('.').str[1]
+                profile_metadata_pd['Object API Name'] = profile_metadata_pd['field'].str.split('.').str[0]
+                profile_metadata_pd = profile_metadata_pd[profile_metadata_pd['Object API Name'].isin(objects)].copy()
+                profile_metadata_pd.drop(columns=['field'], inplace=True)
+
+                # if permission == 'PermissionsEdit':
+                #         profile_metadata_pd = profile_metadata_pd[profile_metadata_pd['editable'] == True].copy()
+                #         profile_metadata_pd.drop(columns=['editable', 'readable'], inplace=True)
+                # elif permission == 'PermissionsRead':
+                #         profile_metadata_pd = profile_metadata_pd[profile_metadata_pd['readable'] == True].copy()
+                #         profile_metadata_pd.drop(columns=['editable', 'readable'], inplace=True)
+                
+                profile_violations = profile_violations.merge(profile_metadata_pd, on='Profile.Name', how='left').copy()
+                profile_editable = profile_violations[profile_violations['editable'] == True].drop(columns=['editable','readable']).copy()
+                profile_editable['Permission'] = 'PermissionsEdit'
+                profile_readable = profile_violations[profile_violations['readable'] == True].drop(columns=['editable','readable']).copy()
+                profile_readable['Permission'] = 'PermissionsRead'
+
+                profile_fls_final = pd.concat([profile_editable, profile_readable])
+                profile_fls_final.columns = ['UserName','Profile','Field API Name','Object API Name','Permission']
+                
+                permission_set_violations = data[data['PermissionSet.IsOwnedByProfile']== False][['User.Username', 'PermissionSet.Label', 'Object API Name', 'FieldPermissions.PermissionsEdit','FieldPermissions.PermissionsRead','Field API Name']]
+                
+                ps_editable = permission_set_violations[permission_set_violations['FieldPermissions.PermissionsEdit'] == True].drop(columns=['FieldPermissions.PermissionsEdit','FieldPermissions.PermissionsRead']).copy()
+                ps_editable['Permission'] = 'PermissionsEdit'
+                ps_readable = permission_set_violations[permission_set_violations['FieldPermissions.PermissionsRead'] == True].drop(columns=['FieldPermissions.PermissionsEdit','FieldPermissions.PermissionsRead']).copy()
+                ps_readable['Permission'] = 'PermissionsRead'
+                
+                ps_fls_final = pd.concat([ps_editable, ps_readable])
+                ps_fls_final.columns = ['UserName','Permission Set','Object API Name','Field API Name', 'Permission']
+                
+                return pd.merge(profile_fls_final, ps_fls_final, on=['UserName','Object API Name', 'Field API Name','Permission'], how='outer')
+        else:
+                return pd.DataFrame(columns=['UserName', 'Profile', 'Field API Name', 'Object API Name','Permission', 'Permission Set'])
+          
+            
     async def retrieve_user_profile_record_type_details(self, username: str, objects: list[str]):
         
         profile_metadata = pd.DataFrame(zeep.helpers.serialize_object(self.metadata_list("Profile")))
@@ -1061,6 +1167,20 @@ class Sf:
         field_describe.columns = columnNames
         return field_describe
 
+    def set_user_password(self, user_id, new_password):
+        result = {}
+        try:
+            result = self.sf_api_call(f'/services/data/{self.api_version}/sobjects/User/{user_id}/password', method='POST', data={'NewPassword': new_password})
+        except Exception as e:
+            message_list = str(e).split(" : ")[1].strip().removeprefix("b'[").removesuffix("]'")
+            result = ast.literal_eval(message_list)
+            result['user_id'] = user_id
+        
+        if len(result) == 0:
+            result = {'user_id': user_id,'message': 'success',  'errorCode': None}
+        return result
+        
+
     def picklist_dataframe_stacked(self,objects: List =['Account','Address_vod__c','Child_Account_vod__c']) -> pd.DataFrame:
         """
         TODO:
@@ -1190,106 +1310,7 @@ class Sf:
         
         return {'permissionSetPermissions': permissionSetPermissions, 'profilePermissions': profilePermissions}
 
-    async def field_permission_user_check(self, objects: list[str], permissions: list[str], users=None) -> pd.DataFrame:
-        # Similar to the field_permission_check function, but this one will return a list of users that have the specified permission on the specified object
-        
-        # this is to avoid having a mutable default argument
-        users = [] if users is None else users
-        
-        object_query = "','".join(objects)
-        permission_query = ' OR '.join([f"({permission} = true)" for permission in permissions])
 
-        relevant_ps_and_profiles = transform_sf_result_set_rec(self.sf.query_all(f"""
-                                                SELECT ParentId,
-                                                        Field,
-                                                        PermissionsEdit,
-                                                        PermissionsRead
-                                                FROM FieldPermissions
-                                                WHERE SObjectType IN ('{object_query}') AND
-                                                ({permission_query})
-                                                """)['records'])
-
-        relevant_ps_and_profiles['Object API Name'] = relevant_ps_and_profiles['FieldPermissions.Field'].apply(lambda row: row.split('.')[0])
-        relevant_ps_and_profiles['Field API Name'] = relevant_ps_and_profiles['FieldPermissions.Field'].apply(lambda row: row.split('.')[1])
-        relevant_ps_and_profiles.drop(columns=['FieldPermissions.Field'], inplace=True)
-
-        user_query = (" AND Assignee.UserName IN ('" + "','".join(users) + "')") if len(users) > 0 else ""
-
-        relevant_ps_and_profiles_list = "','".join(relevant_ps_and_profiles['FieldPermissions.ParentId'].unique().tolist())
-
-        data = transform_sf_result_set_rec(self.sf.query_all(f"""SELECT Assignee.Id, Assignee.Name,Assignee.IsActive, 
-                                                                Assignee.UserName, PermissionSet.Id, 
-                                                                PermissionSet.isOwnedByProfile, PermissionSet.Profile.Name, PermissionSet.Label
-                                                                FROM PermissionSetAssignment
-                                                                WHERE PermissionSetId
-                                                                IN ('{relevant_ps_and_profiles_list}')
-                                                                {user_query} AND Assignee.IsActive = TRUE""")['records'])
-        
-        if len(data) == 0:
-            return pd.DataFrame(columns=['UserName', 'Profile', 'Field API Name', 'Object API Name','Permission', 'Permission Set'])
-        
-        # Retrieve Profile FLS data
-        profiles = data[data['PermissionSet.IsOwnedByProfile']]['Profile.Name'].unique().tolist()
-        profiles = ['Admin' if x == 'System Administrator' else x for x in profiles]
-
-
-        async_metadata_read = async_wrap(self.metadata_read)
-        profile_metadata_tasks = {}
-        profile_metadata_dict = {}
-
-        for profile in profiles:
-                profile_metadata_tasks[profile] = asyncio.create_task(async_metadata_read('Profile', profile))
-
-
-        task_result_list = await asyncio.gather(*profile_metadata_tasks.values())
-
-
-        for profile in profiles:
-                profile_metadata_dict["System Administrator" if profile == 'Admin' else profile] = pd.DataFrame(task_result_list[profiles.index(profile)]['fieldPermissions'][0])
-    
-        data = data.merge(relevant_ps_and_profiles, left_on='PermissionSet.Id', right_on='FieldPermissions.ParentId', how='left').copy()
-        if (len(data) > 0):
-                
-                profile_violations = data[data['PermissionSet.IsOwnedByProfile']== True][['User.Username', 'Profile.Name']].drop_duplicates().copy()
-                # profile_violations = data[data['PermissionSet.IsOwnedByProfile']== True][['User.Username', 'Profile.Name','Field API Name']]
-                profile_metadata_pd = pd.concat(profile_metadata_dict.values(), keys=profile_metadata_dict.keys(), names=['Profile', 'Index']).reset_index(level=1, drop=True).reset_index()
-                profile_metadata_pd.rename(columns={'Profile': 'Profile.Name'}, inplace=True)
-                profile_metadata_pd['Field API Name'] = profile_metadata_pd['field'].str.split('.').str[1]
-                profile_metadata_pd['Object API Name'] = profile_metadata_pd['field'].str.split('.').str[0]
-                profile_metadata_pd = profile_metadata_pd[profile_metadata_pd['Object API Name'].isin(objects)].copy()
-                profile_metadata_pd.drop(columns=['field'], inplace=True)
-
-                # if permission == 'PermissionsEdit':
-                #         profile_metadata_pd = profile_metadata_pd[profile_metadata_pd['editable'] == True].copy()
-                #         profile_metadata_pd.drop(columns=['editable', 'readable'], inplace=True)
-                # elif permission == 'PermissionsRead':
-                #         profile_metadata_pd = profile_metadata_pd[profile_metadata_pd['readable'] == True].copy()
-                #         profile_metadata_pd.drop(columns=['editable', 'readable'], inplace=True)
-                
-                profile_violations = profile_violations.merge(profile_metadata_pd, on='Profile.Name', how='left').copy()
-                profile_editable = profile_violations[profile_violations['editable'] == True].drop(columns=['editable','readable']).copy()
-                profile_editable['Permission'] = 'PermissionsEdit'
-                profile_readable = profile_violations[profile_violations['readable'] == True].drop(columns=['editable','readable']).copy()
-                profile_readable['Permission'] = 'PermissionsRead'
-
-                profile_fls_final = pd.concat([profile_editable, profile_readable])
-                profile_fls_final.columns = ['UserName','Profile','Field API Name','Object API Name','Permission']
-                
-                permission_set_violations = data[data['PermissionSet.IsOwnedByProfile']== False][['User.Username', 'PermissionSet.Label', 'Object API Name', 'FieldPermissions.PermissionsEdit','FieldPermissions.PermissionsRead','Field API Name']]
-                
-                ps_editable = permission_set_violations[permission_set_violations['FieldPermissions.PermissionsEdit'] == True].drop(columns=['FieldPermissions.PermissionsEdit','FieldPermissions.PermissionsRead']).copy()
-                ps_editable['Permission'] = 'PermissionsEdit'
-                ps_readable = permission_set_violations[permission_set_violations['FieldPermissions.PermissionsRead'] == True].drop(columns=['FieldPermissions.PermissionsEdit','FieldPermissions.PermissionsRead']).copy()
-                ps_readable['Permission'] = 'PermissionsRead'
-                
-                ps_fls_final = pd.concat([ps_editable, ps_readable])
-                ps_fls_final.columns = ['UserName','Permission Set','Object API Name','Field API Name', 'Permission']
-                
-                return pd.merge(profile_fls_final, ps_fls_final, on=['UserName','Object API Name', 'Field API Name','Permission'], how='outer')
-        else:
-                return pd.DataFrame(columns=['UserName', 'Profile', 'Field API Name', 'Object API Name','Permission', 'Permission Set'])
-          
-            
     # Retrieves the Page Layout by Profile by Record Type data for all objects in the org
     def retrieve_profile_layout_record_type_matrix_all(self, return_pivot_table = False):
         account_page_layouts_bulk = transform_sf_result_set_rec(self.tooling_query_all("""select LayoutId, 
