@@ -1249,7 +1249,7 @@ class Vn:
         
         """
         result = requests.get(
-            url="https://demo.vdmdev.com/api/v26.0/systemSettings/api.search.geolocation.countries", 
+            url=f"{self.networkURL}/api/v26.0/systemSettings/api.search.geolocation.countries", 
             headers=self.APIheaders).json()
         return result
         
